/* gtkaudiofilebuffer.c: WaveBuffer object that can read data through
 * audiofile.
 * Copyright (c) 2000 David A. Bartold
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <gtk/gtk.h>
#include "gtkaudiofilebuffer.h"
#include "gwavefile.h"


static void gtk_audiofile_buffer_class_init          (GtkAudiofileBufferClass *klass);
static void gtk_audiofile_buffer_init                (GtkAudiofileBuffer      *audiofile);
static guint32 gtk_audiofile_buffer_get_rate         (GtkWaveBuffer      *wavebuffer);
static GWavefileType gtk_audiofile_buffer_get_datatype  (GtkWaveBuffer  *wavebuffer);
static guint32 gtk_audiofile_buffer_get_length       (GtkWaveBuffer      *wavebuffer);
static guint32 gtk_audiofile_buffer_get_num_channels (GtkWaveBuffer      *wavebuffer);
static void gtk_audiofile_buffer_get_samples         (GtkWaveBuffer      *wavebuffer,
                                                      guint32             start,
                                                      guint32             length,
                                                      guint32             channel_mask,
                                                      gpointer            data);



GtkType
gtk_audiofile_buffer_get_type (void)
{
  static GtkType audiofile_type = 0;

  if (!audiofile_type)
    {
      static const GtkTypeInfo audiofile_info =
      {
        "GtkAudiofileBuffer",
        sizeof (GtkAudiofileBuffer),
        sizeof (GtkAudiofileBufferClass),
        (GtkClassInitFunc) gtk_audiofile_buffer_class_init,
        (GtkObjectInitFunc) gtk_audiofile_buffer_init,
        /* reserved_1 */ NULL,
        /* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL
      };

      audiofile_type = gtk_type_unique (GTK_TYPE_WAVE_BUFFER, &audiofile_info);
    }

  return audiofile_type;
}

static void
gtk_audiofile_buffer_finalize (GtkObject *obj)
{
  GtkAudiofileBuffer *audiofile = GTK_AUDIOFILE_BUFFER (obj);

  afCloseFile (audiofile->file);
}

static void
gtk_audiofile_buffer_class_init (GtkAudiofileBufferClass *klass)
{
  GtkObjectClass *object_class;
  GtkWaveBufferClass *wavebuffer_class;

  object_class = (GtkObjectClass *) klass;
  wavebuffer_class = (GtkWaveBufferClass *) klass;

  object_class->finalize = gtk_audiofile_buffer_finalize;

  wavebuffer_class->get_rate = gtk_audiofile_buffer_get_rate;
  wavebuffer_class->get_datatype = gtk_audiofile_buffer_get_datatype;
  wavebuffer_class->get_length = gtk_audiofile_buffer_get_length;
  wavebuffer_class->get_num_channels = gtk_audiofile_buffer_get_num_channels;
  wavebuffer_class->get_samples = gtk_audiofile_buffer_get_samples;
}

static void
gtk_audiofile_buffer_init (GtkAudiofileBuffer *audiofile)
{
  audiofile->file = NULL;
}

static guint32
gtk_audiofile_buffer_get_rate (GtkWaveBuffer *wavebuffer)
{
  GtkAudiofileBuffer *audiofile = GTK_AUDIOFILE_BUFFER (wavebuffer);

  return afGetRate (audiofile->file, AF_DEFAULT_TRACK);
}

static GWavefileType
gtk_audiofile_buffer_get_datatype (GtkWaveBuffer *wavebuffer)
{
  GtkAudiofileBuffer *audiofile = GTK_AUDIOFILE_BUFFER (wavebuffer);
  int fmt, bits;

  afGetSampleFormat (audiofile->file, AF_DEFAULT_TRACK, &fmt, &bits);

  switch (fmt)
    {
      case AF_SAMPFMT_TWOSCOMP:
        if (bits == 8)
          return G_WAVEFILE_TYPE_S8;
        else if (bits == 16)
          return G_WAVEFILE_TYPE_S16;
        else if (bits == 32)
          return G_WAVEFILE_TYPE_S32;
        break;

      case AF_SAMPFMT_UNSIGNED:
        if (bits == 8)
          return G_WAVEFILE_TYPE_U8;
        else if (bits == 16)
          return G_WAVEFILE_TYPE_U16;
        else if (bits == 32)
          return G_WAVEFILE_TYPE_U32;
        break;
    }

  return G_WAVEFILE_TYPE_NULL;
}

static guint32
gtk_audiofile_buffer_get_length (GtkWaveBuffer *wavebuffer)
{
  GtkAudiofileBuffer *audiofile = GTK_AUDIOFILE_BUFFER (wavebuffer);

  return afGetFrameCount (audiofile->file, AF_DEFAULT_TRACK);
}

static guint32
gtk_audiofile_buffer_get_num_channels (GtkWaveBuffer *wavebuffer)
{
  GtkAudiofileBuffer *audiofile = GTK_AUDIOFILE_BUFFER (wavebuffer);

  return afGetChannels (audiofile->file, AF_DEFAULT_TRACK);
}

static void
gtk_audiofile_buffer_get_samples (GtkWaveBuffer *wavebuffer,
                                  guint32        start,
                                  guint32        length,
                                  guint32        channel_mask,
                                  gpointer       data)
{
  GtkAudiofileBuffer *audiofile = GTK_AUDIOFILE_BUFFER (wavebuffer);
  guint32 width, n_channels, mask;

  width = g_wavefile_type_width (gtk_audiofile_buffer_get_datatype (wavebuffer));
  n_channels = afGetChannels (audiofile->file, AF_DEFAULT_TRACK);

  afSeekFrame (audiofile->file, AF_DEFAULT_TRACK, start);

  mask = (1 << (n_channels + 1)) - 1;

  if ((mask & channel_mask) == mask)
    afReadFrames (audiofile->file, AF_DEFAULT_TRACK, data, length);
  else
    {
      guint8 *buf, *bufpos;
      guint32 pos, size, i, j;

      buf = g_malloc (width * n_channels * length);
      pos = 0;
      while (pos < length)
        {
          size = length;
          if (size > 1024)
            size = 1024;

          afReadFrames (audiofile->file, AF_DEFAULT_TRACK, buf, size);

          bufpos = buf;
          for (i = 0; i < size; i++)
            for (j = 0; j < n_channels; j++)
              {
                if (channel_mask & (1 << j))
                  {
                    memcpy (data, bufpos, width);
                    data += width;
                  }

                bufpos += width;
              }

          pos += size;
        }
      g_free (buf);
    }
}

GtkObject *
gtk_audiofile_buffer_new (gchar *filename)
{
  GtkAudiofileBuffer *wavebuffer;
  AFfilehandle h;

  h = afOpenFile (filename, "r", NULL);

  wavebuffer = gtk_type_new (GTK_TYPE_AUDIOFILE_BUFFER);
  wavebuffer->file = h;

  return GTK_OBJECT (wavebuffer);
}
