/* gtkmemorywavebuffer.c: class that stores data in memory
 * Copyright (c) 2000 David A. Bartold
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <gtk/gtk.h>
#include "gtkmemorywavebuffer.h"


static void gtk_memory_wave_buffer_class_init (GtkMemoryWaveBufferClass *klass);
static void gtk_memory_wave_buffer_init       (GtkMemoryWaveBuffer      *memory);
static gboolean gtk_memory_wave_buffer_is_editable     (GtkWaveBuffer *wavebuffer);
static guint32 gtk_memory_wave_buffer_get_rate         (GtkWaveBuffer *wavebuffer);
static GWavefileType gtk_memory_wave_buffer_get_datatype  (GtkWaveBuffer *wavebuffer);
static guint32 gtk_memory_wave_buffer_get_length       (GtkWaveBuffer *wavebuffer);
static guint32 gtk_memory_wave_buffer_get_num_channels (GtkWaveBuffer *wavebuffer);
static void gtk_memory_wave_buffer_get_samples         (GtkWaveBuffer *wavebuffer,
                                                        guint32 start,
                                                        guint32 length,
                                                        guint32 channel_mask,
                                                        gpointer data);
static gint gtk_memory_wave_buffer_set_samples         (GtkEditableWaveBuffer *editable,
                                                        guint32 start,
                                                        guint32 length,
                                                        guint32 channel_mask,
                                                        gpointer data);
static gint gtk_memory_wave_buffer_insert              (GtkEditableWaveBuffer *editable,
                                                        guint32 start,
                                                        guint32 length);
static gint gtk_memory_wave_buffer_delete              (GtkEditableWaveBuffer *editable,
                                                        guint32 start,
                                                        guint32 length);


GtkType
gtk_memory_wave_buffer_get_type (void)
{
  static GtkType memory_type = 0;

  if (!memory_type)
    {
      static const GtkTypeInfo memory_info =
      {
        "GtkMemoryWaveBuffer",
        sizeof (GtkMemoryWaveBuffer),
        sizeof (GtkMemoryWaveBufferClass),
        (GtkClassInitFunc) gtk_memory_wave_buffer_class_init,
        (GtkObjectInitFunc) gtk_memory_wave_buffer_init,
        /* reserved_1 */ NULL,
        /* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL
      };

      memory_type = gtk_type_unique (GTK_TYPE_EDITABLE_WAVE_BUFFER, &memory_info);
    }

  return memory_type;
}


static void
gtk_memory_wave_buffer_finalize (GtkObject *obj)
{
  GtkMemoryWaveBuffer *memory = GTK_MEMORY_WAVE_BUFFER (obj);

  g_free (memory->data);
}


static void
gtk_memory_wave_buffer_class_init (GtkMemoryWaveBufferClass *klass)
{
  GtkObjectClass *object_class;
  GtkWaveBufferClass *wavebuffer_class;
  GtkEditableWaveBufferClass *editable_class;

  object_class = (GtkObjectClass *) klass;
  wavebuffer_class = (GtkWaveBufferClass *) klass;
  editable_class = (GtkEditableWaveBufferClass *) klass;

  object_class->finalize = gtk_memory_wave_buffer_finalize;

  wavebuffer_class->get_rate = gtk_memory_wave_buffer_get_rate;
  wavebuffer_class->get_datatype = gtk_memory_wave_buffer_get_datatype;
  wavebuffer_class->get_length = gtk_memory_wave_buffer_get_length;
  wavebuffer_class->get_num_channels = gtk_memory_wave_buffer_get_num_channels;
  wavebuffer_class->get_samples = gtk_memory_wave_buffer_get_samples;
  editable_class->set_samples = gtk_memory_wave_buffer_set_samples;
  editable_class->insert = gtk_memory_wave_buffer_insert;
  editable_class->delete = gtk_memory_wave_buffer_delete;
}


static void
gtk_memory_wave_buffer_init (GtkMemoryWaveBuffer *memory)
{
  memory->n_channels = 0;
  memory->d_type = G_WAVEFILE_TYPE_NULL;
  memory->length = 0;
  memory->rate = 0;
  memory->data = NULL;
}


static guint32
gtk_memory_wave_buffer_get_rate (GtkWaveBuffer *wavebuffer)
{
  GtkMemoryWaveBuffer *memory = GTK_MEMORY_WAVE_BUFFER (wavebuffer);

  return memory->rate;
}


static GWavefileType
gtk_memory_wave_buffer_get_datatype (GtkWaveBuffer *wavebuffer)
{
  GtkMemoryWaveBuffer *memory = GTK_MEMORY_WAVE_BUFFER (wavebuffer);

  return memory->d_type;
}


static guint32
gtk_memory_wave_buffer_get_length (GtkWaveBuffer *wavebuffer)
{
  GtkMemoryWaveBuffer *memory = GTK_MEMORY_WAVE_BUFFER (wavebuffer);

  return memory->length;
}


static guint32
gtk_memory_wave_buffer_get_num_channels (GtkWaveBuffer *wavebuffer)
{
  GtkMemoryWaveBuffer *memory = GTK_MEMORY_WAVE_BUFFER (wavebuffer);

  return memory->n_channels;
}


static void
gtk_memory_wave_buffer_get_samples (GtkWaveBuffer *wavebuffer,
                                    guint32  start,
                                    guint32  length,
                                    guint32  channel_mask,
                                    gpointer data)
{
  GtkMemoryWaveBuffer *memory = GTK_MEMORY_WAVE_BUFFER (wavebuffer);
  guint32 i, j, width;
  gpointer ptr;

  width = g_wavefile_type_width (memory->d_type);
  ptr = (gpointer) &((guint8*) (memory->data))[start * width * memory->n_channels];

  for (i = 0; i < length; i++)
    for (j = 0; j < memory->n_channels; j++)
      {
        if (channel_mask & (1 << j))
          {
            memcpy (data, ptr, width);
            data = (gpointer) (((guint8*) data) + width);
          }

        ptr = (gpointer) (((guint8*) ptr) + width);
      }
}


static gint
gtk_memory_wave_buffer_set_samples (GtkEditableWaveBuffer *editable,
                                    guint32  start,
                                    guint32  length,
                                    guint32  channel_mask,
                                    gpointer data)
{
  GtkMemoryWaveBuffer *memory = GTK_MEMORY_WAVE_BUFFER (editable);
  guint32 i, j, width;
  gpointer ptr;

  width = g_wavefile_type_width (memory->d_type);
  ptr = (gpointer) &((guint8*) (memory->data))[start * width * memory->n_channels];

  for (i = 0; i < length; i++)
    for (j = 0; j < memory->n_channels; j++)
      {
        if (channel_mask & (1 << j))
          {
            memcpy (ptr, data, width);
            data = (gpointer) (((guint8*) data) + width);
          }

        ptr = (gpointer) (((guint8*) ptr) + width);
      }

  gtk_editable_wave_buffer_queue_modified (editable, start, length);

  return 0;
}


static gint
gtk_memory_wave_buffer_insert (GtkEditableWaveBuffer *editable,
                               guint32 start,
                               guint32 length)
{
  GtkMemoryWaveBuffer *memory = GTK_MEMORY_WAVE_BUFFER (editable);
  guint32 width;
  gpointer *data;
  GRange range;

  if (start > memory->length)
    start = memory->length;

  width = g_wavefile_type_width (memory->d_type) * memory->n_channels;

  data = g_malloc ((memory->length + length) * width);
  g_free (memory->data);

  memmove (data, memory->data, start * width);
  memmove (&((guint8*) data)[(start + length) * width], &((guint8*) memory->data)[start * width], (memory->length - start) * width);

  memory->data = data;
  memory->length += length;

  g_range_set (&range, start, length);
  gtk_signal_emit (GTK_OBJECT (editable), insert_data_signal, &range);

  gtk_editable_wave_buffer_queue_modified (editable, start + length, memory->length - (start + length));

  return 0;
}


static gint
gtk_memory_wave_buffer_delete (GtkEditableWaveBuffer *editable,
                               guint32 start,
                               guint32 length)
{
  GtkMemoryWaveBuffer *memory = GTK_MEMORY_WAVE_BUFFER (editable);
  gpointer data;
  guint32 width;
  guint32 orig_length;
  GRange range;

  width = g_wavefile_type_width (memory->d_type) * memory->n_channels;
  orig_length = memory->length;
  memory->length = orig_length - length;

  g_range_set (&range, start, length);
  gtk_signal_emit (GTK_OBJECT (editable), delete_data_signal, &range);
  gtk_editable_wave_buffer_queue_modified (editable, start, orig_length);

  start *= width;
  length *= width;
  orig_length *= width;

  data = g_malloc (memory->length * width);
  memmove (data, memory->data, start);
  memmove (&((guint8*) data)[start], &((guint8*) memory->data)[start + length], orig_length - length - start);

  g_free (memory->data);
  memory->data = data;

  return 0;
}


GtkObject *
gtk_memory_wave_buffer_new (guint16       n_channels,
                            GWavefileType d_type,
                            guint32       rate,
                            guint32       length)
{
  GtkMemoryWaveBuffer *wavebuffer;

  wavebuffer = gtk_type_new (GTK_TYPE_MEMORY_WAVE_BUFFER);
  wavebuffer->n_channels = n_channels;
  wavebuffer->d_type = d_type;
  wavebuffer->rate = rate;
  wavebuffer->length = length;

  wavebuffer->data = g_malloc (n_channels * g_wavefile_type_width (d_type) * length);

  g_wavefile_type_clear (n_channels, length, d_type, wavebuffer->data);

  return GTK_OBJECT (wavebuffer);
}
