/* gtktempfilewavebuffer.c: class that stores data in a tempfile
 * Copyright (c) 2000 David A. Bartold
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <gtk/gtk.h>
#include "gtktempfilewavebuffer.h"


static void gtk_tempfile_wave_buffer_class_init (GtkTempfileWaveBufferClass *klass);
static void gtk_tempfile_wave_buffer_init       (GtkTempfileWaveBuffer      *tempfile);
static gboolean gtk_tempfile_wave_buffer_is_editable     (GtkWaveBuffer *wavebuffer);
static guint32 gtk_tempfile_wave_buffer_get_rate         (GtkWaveBuffer *wavebuffer);
static GWavefileType gtk_tempfile_wave_buffer_get_datatype  (GtkWaveBuffer *wavebuffer);
static guint32 gtk_tempfile_wave_buffer_get_length       (GtkWaveBuffer *wavebuffer);
static guint32 gtk_tempfile_wave_buffer_get_num_channels (GtkWaveBuffer *wavebuffer);
static void gtk_tempfile_wave_buffer_get_samples         (GtkWaveBuffer *wavebuffer,
                                                          guint32 start,
                                                          guint32 length,
                                                          guint32 channel_mask,
                                                          gpointer data);
static gint gtk_tempfile_wave_buffer_set_samples         (GtkEditableWaveBuffer *editable,
                                                          guint32 start,
                                                          guint32 length,
                                                          guint32 channel_mask,
                                                          gpointer data);
static gint gtk_tempfile_wave_buffer_insert              (GtkEditableWaveBuffer *editable,
                                                          guint32 start,
                                                          guint32 length);
static gint gtk_tempfile_wave_buffer_delete              (GtkEditableWaveBuffer *editable,
                                                      guint32 start,
                                                      guint32 length);



GtkType
gtk_tempfile_wave_buffer_get_type (void)
{
  static GtkType tempfile_type = 0;

  if (!tempfile_type)
    {
      static const GtkTypeInfo tempfile_info =
      {
        "GtkTempfileWaveBuffer",
        sizeof (GtkTempfileWaveBuffer),
        sizeof (GtkTempfileWaveBufferClass),
        (GtkClassInitFunc) gtk_tempfile_wave_buffer_class_init,
        (GtkObjectInitFunc) gtk_tempfile_wave_buffer_init,
        /* reserved_1 */ NULL,
        /* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL
      };

      tempfile_type = gtk_type_unique (GTK_TYPE_EDITABLE_WAVE_BUFFER, &tempfile_info);
    }

  return tempfile_type;
}

static void
gtk_tempfile_wave_buffer_finalize (GtkObject *obj)
{
  GtkTempfileWaveBuffer *tempfile = GTK_TEMPFILE_WAVE_BUFFER (obj);

  g_print ("Finalize called.\n");
  fclose (tempfile->file);

  if (tempfile->in_use != NULL)
    g_mutex_free (tempfile->in_use);
}

static void
gtk_tempfile_wave_buffer_class_init (GtkTempfileWaveBufferClass *klass)
{
  GtkObjectClass *object_class;
  GtkWaveBufferClass *wavebuffer_class;
  GtkEditableWaveBufferClass *editable_class;

  object_class = (GtkObjectClass *) klass;
  wavebuffer_class = (GtkWaveBufferClass *) klass;
  editable_class = (GtkEditableWaveBufferClass *) klass;

  object_class->finalize = gtk_tempfile_wave_buffer_finalize;

  wavebuffer_class->get_rate = gtk_tempfile_wave_buffer_get_rate;
  wavebuffer_class->get_datatype = gtk_tempfile_wave_buffer_get_datatype;
  wavebuffer_class->get_length = gtk_tempfile_wave_buffer_get_length;
  wavebuffer_class->get_num_channels = gtk_tempfile_wave_buffer_get_num_channels;
  wavebuffer_class->get_samples = gtk_tempfile_wave_buffer_get_samples;
  editable_class->set_samples = gtk_tempfile_wave_buffer_set_samples;
  editable_class->insert = gtk_tempfile_wave_buffer_insert;
  editable_class->delete = gtk_tempfile_wave_buffer_delete;
}

static void
gtk_tempfile_wave_buffer_init (GtkTempfileWaveBuffer *tempfile)
{
  tempfile->n_channels = 0;
  tempfile->d_type = G_WAVEFILE_TYPE_NULL;
  tempfile->length = 0;
  tempfile->rate = 0;
  tempfile->file = tmpfile ();

  if (g_threads_got_initialized)
    tempfile->in_use = g_mutex_new ();
  else
    tempfile->in_use = NULL;
}

static guint32
gtk_tempfile_wave_buffer_get_rate (GtkWaveBuffer *wavebuffer)
{
  GtkTempfileWaveBuffer *tempfile = GTK_TEMPFILE_WAVE_BUFFER (wavebuffer);

  return tempfile->rate;
}

static GWavefileType
gtk_tempfile_wave_buffer_get_datatype (GtkWaveBuffer *wavebuffer)
{
  GtkTempfileWaveBuffer *tempfile = GTK_TEMPFILE_WAVE_BUFFER (wavebuffer);

  return tempfile->d_type;
}

static guint32
gtk_tempfile_wave_buffer_get_length (GtkWaveBuffer *wavebuffer)
{
  GtkTempfileWaveBuffer *tempfile = GTK_TEMPFILE_WAVE_BUFFER (wavebuffer);

  return tempfile->length;
}

static guint32
gtk_tempfile_wave_buffer_get_num_channels (GtkWaveBuffer *wavebuffer)
{
  GtkTempfileWaveBuffer *tempfile = GTK_TEMPFILE_WAVE_BUFFER (wavebuffer);

  return tempfile->n_channels;
}

static void
gtk_tempfile_wave_buffer_get_samples (GtkWaveBuffer *wavebuffer,
                                      guint32 start,
                                      guint32 length,
                                      guint32 channel_mask,
                                      gpointer data)
{
  GtkTempfileWaveBuffer *tempfile = GTK_TEMPFILE_WAVE_BUFFER (wavebuffer);
  guint32 mask, i, j, width;

  if (tempfile->in_use != NULL)
    g_mutex_lock (tempfile->in_use);

  width = g_wavefile_type_width (tempfile->d_type);
  fseek (tempfile->file, start * width * tempfile->n_channels, SEEK_SET);

  mask = (1 << (tempfile->n_channels + 1)) - 1;

  if ((mask & channel_mask) == mask)
    fread (data, width * tempfile->n_channels, length, tempfile->file);
  else
    {
      guint8 junk[width * tempfile->n_channels];

      for (i = 0; i < length; i++)
        for (j = 0; j < tempfile->n_channels; j++)
          if (channel_mask & (1 << j))
            {
              fread (data, width, 1, tempfile->file);
              data = (gpointer) (((guint8*) data) + width);
            }
          else
            fread (junk, width, 1, tempfile->file);
    }

  if (tempfile->in_use != NULL)
    g_mutex_unlock (tempfile->in_use);
}

static gint
gtk_tempfile_wave_buffer_set_samples (GtkEditableWaveBuffer *editable,
                                      guint32 start,
                                      guint32 length,
                                      guint32 channel_mask,
                                      gpointer data)
{
  GtkTempfileWaveBuffer *tempfile = GTK_TEMPFILE_WAVE_BUFFER (editable);
  guint32 mask, i, j, width;

  if (tempfile->in_use != NULL)
    g_mutex_lock (tempfile->in_use);

  width = g_wavefile_type_width (tempfile->d_type);
  fseek (tempfile->file, start * width * tempfile->n_channels, SEEK_SET);

  mask = (1 << (tempfile->n_channels + 1)) - 1;

  if ((mask & channel_mask) == mask)
    fwrite (data, width * tempfile->n_channels, length, tempfile->file);
  else
    {
      guint8 junk[width * tempfile->n_channels];

      for (i = 0; i < length; i++)
        for (j = 0; j < tempfile->n_channels; j++)
          if (channel_mask & (1 << j))
            {
              fwrite (data, width, 1, tempfile->file);
              data = (gpointer) (((guint8*) data) + width);
            }
          else
            fread (junk, width, 1, tempfile->file);
    }

  gtk_editable_wave_buffer_queue_modified (editable, start, length);

  if (tempfile->in_use != NULL)
    g_mutex_unlock (tempfile->in_use);

  return 0;
}

static void
gtk_tempfile_move_data (GtkTempfileWaveBuffer *tempfile,
                        guint32 from_pos,
                        guint32 to_pos,
                        guint32 length)
{
  guint8 block[4096];
  gint32 width, pos, size;

  if (from_pos == to_pos)
    return;

  width = g_wavefile_type_width (tempfile->d_type) * tempfile->n_channels;

  length *= width;
  from_pos *= width;
  to_pos *= width;

  if (to_pos < from_pos)
    {
      pos = 0;
      while (pos < length)
        {
          size = MIN (4096, length - pos);

          fseek (tempfile->file, from_pos + pos, SEEK_SET);
          fread (block, 1, size, tempfile->file);

          fseek (tempfile->file, to_pos + pos, SEEK_SET);
          fwrite (block, 1, size, tempfile->file);

          pos += size;
        }
    }
  else
    {
      pos = length;
      while (pos > 0)
        {
          size = MIN (4096, pos);

          fseek (tempfile->file, from_pos + pos, SEEK_SET);
          fread (block, 1, size, tempfile->file);

          fseek (tempfile->file, to_pos + pos, SEEK_SET);
          fwrite (block, 1, size, tempfile->file);

          pos -= size;
        }
    }
}

static gint
gtk_tempfile_wave_buffer_insert (GtkEditableWaveBuffer *editable,
                                 guint32 start,
                                 guint32 ins_length)
{
  GtkTempfileWaveBuffer *tempfile = GTK_TEMPFILE_WAVE_BUFFER (editable);
  gint32 length;
  GRange range;

  length = gtk_wave_buffer_get_length (GTK_WAVE_BUFFER (editable));

  if (tempfile->in_use != NULL)
    g_mutex_lock (tempfile->in_use);

  gtk_tempfile_move_data (tempfile, start, start + ins_length, length - start);
  tempfile->length += ins_length;

  g_range_set (&range, start, length);
  gtk_signal_emit (GTK_OBJECT (editable), insert_data_signal, &range);

  gtk_editable_wave_buffer_queue_modified (editable, start, tempfile->length - start);

  if (tempfile->in_use != NULL)
    g_mutex_unlock (tempfile->in_use);

  return ferror (tempfile->file);
}

static gint
gtk_tempfile_wave_buffer_delete (GtkEditableWaveBuffer *editable,
                                 guint32 start,
                                 guint32 length)
{
  GtkTempfileWaveBuffer *tempfile = GTK_TEMPFILE_WAVE_BUFFER (editable);
  guint32 orig_length;
  GRange range;

  if (tempfile->in_use != NULL)
    g_mutex_lock (tempfile->in_use);

  orig_length = tempfile->length;
  gtk_tempfile_move_data (tempfile, start + length, start, orig_length - length);

  tempfile->length = orig_length - length;
  ftruncate (fileno (tempfile->file), tempfile->length);

  g_range_set (&range, start, length);
  gtk_signal_emit (GTK_OBJECT (editable), delete_data_signal, &range);

  gtk_editable_wave_buffer_queue_modified (editable, start, orig_length - start);

  if (tempfile->in_use != NULL)
    g_mutex_unlock (tempfile->in_use);

  return ferror (tempfile->file);
}

GtkObject *
gtk_tempfile_wave_buffer_new (guint16       n_channels,
                              GWavefileType d_type,
                              guint32       rate)
{
  GtkTempfileWaveBuffer *wavebuffer;

  wavebuffer = gtk_type_new (GTK_TYPE_TEMPFILE_WAVE_BUFFER);
  wavebuffer->n_channels = n_channels;
  wavebuffer->d_type = d_type;
  wavebuffer->rate = rate;
  wavebuffer->length = 0;

  return GTK_OBJECT (wavebuffer);
}
