/* gtkwavefilebuffer.c: class to read wave data directly from a .WAV file
 * Copyright (c) 2000 David A. Bartold
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <gtk/gtk.h>
#include "gtkwavefilebuffer.h"


static void gtk_wavefile_buffer_class_init          (GtkWavefileBufferClass *klass);
static void gtk_wavefile_buffer_init                (GtkWavefileBuffer      *fast);
static guint32 gtk_wavefile_buffer_get_rate         (GtkWaveBuffer          *wavebuffer);
static GWavefileType gtk_wavefile_buffer_get_datatype  (GtkWaveBuffer   *wavebuffer);
static guint32 gtk_wavefile_buffer_get_length       (GtkWaveBuffer          *wavebuffer);
static guint32 gtk_wavefile_buffer_get_num_channels (GtkWaveBuffer          *wavebuffer);
static void gtk_wavefile_buffer_get_samples         (GtkWaveBuffer          *wavebuffer,
                                                     guint32                 start,
                                                     guint32                 length,
                                                     guint32                 channel_mask,
                                                     gpointer                data);



GtkType
gtk_wavefile_buffer_get_type (void)
{
  static GtkType wave_type = 0;

  if (!wave_type)
    {
      static const GtkTypeInfo wave_info =
      {
        "GtkWavefileBuffer",
        sizeof (GtkWavefileBuffer),
        sizeof (GtkWavefileBufferClass),
        (GtkClassInitFunc) gtk_wavefile_buffer_class_init,
        (GtkObjectInitFunc) gtk_wavefile_buffer_init,
        /* reserved_1 */ NULL,
        /* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL
      };

      wave_type = gtk_type_unique (GTK_TYPE_WAVE_BUFFER, &wave_info);
    }

  return wave_type;
}

static void
gtk_wavefile_buffer_finalize (GtkObject *obj)
{
  GtkWavefileBuffer *fast = GTK_WAVEFILE_BUFFER (obj);

  g_wavefile_close (fast->data);
}

static void
gtk_wavefile_buffer_class_init (GtkWavefileBufferClass *klass)
{
  GtkObjectClass *object_class;
  GtkWaveBufferClass *wavebuffer_class;

  object_class = (GtkObjectClass *) klass;
  wavebuffer_class = (GtkWaveBufferClass *) klass;

  object_class->finalize = gtk_wavefile_buffer_finalize;

  wavebuffer_class->get_rate = gtk_wavefile_buffer_get_rate;
  wavebuffer_class->get_datatype = gtk_wavefile_buffer_get_datatype;
  wavebuffer_class->get_length = gtk_wavefile_buffer_get_length;
  wavebuffer_class->get_num_channels = gtk_wavefile_buffer_get_num_channels;
  wavebuffer_class->get_samples = gtk_wavefile_buffer_get_samples;
}

static void
gtk_wavefile_buffer_init (GtkWavefileBuffer *fast)
{
  fast->data = NULL;
}

static guint32
gtk_wavefile_buffer_get_rate (GtkWaveBuffer *wavebuffer)
{
  GtkWavefileBuffer *fast = GTK_WAVEFILE_BUFFER (wavebuffer);

  return g_wavefile_get_rate (fast->data);
}

static GWavefileType
gtk_wavefile_buffer_get_datatype (GtkWaveBuffer *wavebuffer)
{
  GtkWavefileBuffer *fast = GTK_WAVEFILE_BUFFER (wavebuffer);

  return g_wavefile_get_datatype (fast->data);
}

static guint32
gtk_wavefile_buffer_get_length (GtkWaveBuffer *wavebuffer)
{
  GtkWavefileBuffer *fast = GTK_WAVEFILE_BUFFER (wavebuffer);

  return g_wavefile_get_length (fast->data);
}

static guint32
gtk_wavefile_buffer_get_num_channels (GtkWaveBuffer *wavebuffer)
{
  GtkWavefileBuffer *fast = GTK_WAVEFILE_BUFFER (wavebuffer);

  return g_wavefile_get_num_channels (fast->data);
}

static void
gtk_wavefile_buffer_get_samples (GtkWaveBuffer *wavebuffer,
                                 guint32  start,
                                 guint32  length,
                                 guint32  channel_mask,
                                 gpointer data)
{
  GtkWavefileBuffer *fast = GTK_WAVEFILE_BUFFER (wavebuffer);
  guint32 n_channels, mask;

  n_channels = gtk_wavefile_buffer_get_num_channels (wavebuffer);

  mask = (1 << (n_channels + 1)) - 1;

  if ((mask & channel_mask) == mask)
    g_wavefile_read (fast->data, start, length, data);
  else
    {
      guint8 *buf, *bufpos;
      guint32 width, pos, size, i, j;

      width = g_wavefile_type_width (gtk_wavefile_buffer_get_datatype (wavebuffer));
      buf = g_malloc (width * n_channels * length);
      pos = 0;
      while (pos < length)
        {
          size = length;
          if (size > 1024)
            size = 1024;

          g_wavefile_read (fast->data, start + pos, size, buf);

          bufpos = buf;
          for (i = 0; i < size; i++)
            for (j = 0; j < n_channels; j++)
              {
                if (channel_mask & (1 << j))
                  {
                    memcpy (data, bufpos, width);
                    data += width;
                  }

                bufpos += width;
              }

          pos += size;
        }
      g_free (buf);
    }
}

GtkObject *
gtk_wavefile_buffer_new (gchar *filename)
{
  GtkWavefileBuffer *wavebuffer;
  GWavefile *file;

  file = g_wavefile_open (filename);

  if (file == NULL)
    return NULL;

  wavebuffer = gtk_type_new (GTK_TYPE_WAVEFILE_BUFFER);
  wavebuffer->data = file;

  return GTK_OBJECT (wavebuffer);
}
