/*
 *   mpl - Macro Processing Language
 *   Copyright (C) 1997  David A. Bartold
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <ctype.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define TYP_NULL	0
#define TYP_FILE	1
#define TYP_STR		2

#define STO_NULL	0
#define STO_ENDPARAM	1

#define NOD_NULL	0
#define NOD_VAR		1
#define NOD_FUNC	2

#define M_NULL		0
#define M_FOR		1
#define M_ADD		2
#define M_SUB		3
#define M_MUL		4
#define M_DIV		5
#define M_MOD		6
#define M_UP		7
#define M_LOW		8
#define M_LEN		9
#define M_DNL		10
#define M_HEX		11
#define M_HEXBYTE	12
#define M_IF		13
#define M_OR		14
#define M_AND		15
#define M_EQ		16
#define M_NE		17
#define M_GT		18
#define M_GE		19
#define M_LT		20
#define M_LE		21
#define M_XOR		22
#define M_NOT		23
#define M_NL		24
#define M_RAND		25
#define M_MKLIST	26
#define M_SUBSTR	27

struct Macro
{
	int num;
	char *name;
	char *formal;
};

Macro macros[] =
{
	{ M_FOR, "for", "tiiie" },
	{ M_ADD, "add", "ii+" },
	{ M_SUB, "sub", "ii+" },
	{ M_MUL, "mul", "ii+" },
	{ M_DIV, "div", "ii+" },
	{ M_MOD, "mod", "ii+" },
	{ M_OR, "or", "ii+" },
	{ M_UP, "toupper", "i" },
	{ M_LOW, "tolower", "i" },
	{ M_LEN, "strlen", "i" },
	{ M_DNL, "dnl", "" },
	{ M_HEX, "hex", "i" },
	{ M_HEXBYTE, "hexbyte", "i" },
	{ M_IF, "if", "ie?" },
	{ M_OR, "or", "ee+" },
	{ M_AND, "and", "ee+" },
	{ M_EQ, "eq", "ii" },
	{ M_NE, "ne", "ii" },
	{ M_GT, "gt", "ii" },
	{ M_GE, "ge", "ii" },
	{ M_LT, "lt", "ii" },
	{ M_LE, "le", "ii" },
	{ M_XOR, "xor", "ii" },
	{ M_NOT, "not", "i" },
	{ M_NL, "nl", "" },
	{ M_RAND, "rand", "" },
	{ M_MKLIST, "mklist", "i+" },
	{ M_SUBSTR, "substr", "iii" },
	{ M_NULL, "<NULL>", NULL }
};

inline int sgn( int i )
{
	if( i<0 )
	{
		return -1;
	}
	else if( i>0 )
	{
		return 1;
	}

	return 0;
}

class String
{
	int len;
	int size;
	char *str;

	void resize( int newsize );

public:
	char operator[]( int a ) { return a<len ? str[ a ] : 0; }
	String& operator+=( char c );
	String& operator+=( String& s );
	String& operator=( char *s );
	String& operator=( String& s );
	String& operator=( int i );
	int operator==( char *s );
	int operator==( String& s );
	int operator!=( char *s ) { return !(*this==s); }
	int operator!=( String& s ) { return !(*this==s); }
	void operator<<( String& s );
	int hash();
	int isNumber();
	void set( char *s, int l );
	void setHex( int i );
	void setHexByte( int i );
	int smartCompare( String& s );
	int stringCompare( String& s );
	void substr( String& s, int first, int last );
	int toInt();
	void toLower();
	void toUpper();
	void write( FILE *out );
	int getLength() { return len; }

	String();
	~String();
};

void String::resize( int newsize )
{
	char *n;

	if( newsize>size )
	{
		n = new char[ newsize ];
		if( n==NULL )
		{
			fprintf( stderr, "You're screwed!\n" );
			exit( 69 );
		}

		memcpy( n, str, len );
		size = newsize;
		delete str;
		str = n;
	}
}

String& String::operator+=( char c )
{
	if( len>=size )
	{
		resize( size+128 );
	}

	str[ len++ ] = c;

	return *this;
}

String& String::operator+=( String& s )
{
	if( len+s.len>size )
	{
		resize( len+s.len+128 );
	}

	memcpy( &str[ len ], s.str, s.len );
	len = len + s.len;

	return *this;
}

String& String::operator=( char *s )
{
	set( s, strlen( s ) );

	return *this;
}

String& String::operator=( String& s )
{
	set( s.str, s.len );

	return *this;
}

String& String::operator=( int i )
{
	char num[ 20 ];

	sprintf( num, "%i", i );
	*this = num;

	return *this;
}

int String::operator==( char *s )
{
	if( len!=strlen( s ) )
	{
		return 0;
	}

	return memcmp( s, str, len )==0;
}

int String::operator==( String& s )
{
	if( len!=s.len )
	{
		return 0;
	}

	return memcmp( str, s.str, len )==0;
}

void String::operator<<( String& s )
{
	delete str;

	len = s.len;
	str = s.str;
	size = s.size;

	s.len = 0;
	s.str = NULL;
	s.size = 0;
}

int String::hash()
{
	int g;
	unsigned sum;

	sum = 0;
	for( g=0; g<len; g++ )
	{
		sum = (sum+str[ g ]+3359)*1787;
	}

	return sum%1009;
}

int String::isNumber()
{
	int g;

	for( g=0; g<len && str[ g ]<=' '; g++ )
	{
	}

	if( g<len && str[ g ]=='-' )
	{
		g++;
		if( g>=len || str[ g ]<'0' || str[ g ]>'9' )
		{
			return 0;
		}
	}

	for( ; g<len && str[ g ]>='0' && str[ g ]<='9'; g++ )
	{
	}

	for( ; g<len && str[ g ]<=' '; g++ )
	{
	}

	return g==len;
}

void String::set( char *s, int l )
{
	delete str;
	str = NULL;
	len = 0;
	size = 0;

	if( l!=0 )
	{
		resize( l+128 );
		memcpy( str, s, l );
		len = l;
	}
}

void String::setHex( int i )
{
	char num[ 20 ];

	sprintf( num, "%x", i );
	*this =	num;
}

void String::setHexByte( int i )
{
	char num[ 8 ];

	sprintf( num, "%02x", i&0xff );
	*this =	num;
}

int String::smartCompare( String& s )
{
	if( isNumber() && s.isNumber() )
	{
		return sgn( toInt()-s.toInt() );
	}

	return stringCompare( s );
}

int String::stringCompare( String& s )
{
	int temp;
	int g;

	temp = len<s.len ? len : s.len;
	for( g=0; g<temp; g++ )
	{
		if( str[ g ]!=s.str[ g ] )
		{
			return sgn( str[ g ]-s.str[ g ] );
		}
	}

	return sgn( len-s.len );
}

void String::substr( String& s, int first, int last )
{
	if( last<first )
	{
		*this = "";
		return;
	}
	if( first<0 )
	{
		first = 0;
	}
	if( first>=s.len )
	{
		first = s.len-1;
	}
	if( last<0 )
	{
		last = 0;
	}
	if( last>=s.len )
	{
		last = s.len-1;
	}

	set( &s.str[ first ], last-first+1 );
}

int String::toInt()
{
	int num, neg, g;

	for( g=0; g<len && str[ g ]<=' '; g++ )
	{
	}

	if( g<len && str[ g ]=='-' )
	{
		neg = 1;
		g++;
	}
	else
	{
		neg = 0;
	}

	for( num=0; g<len && str[ g ]>='0' && str[ g ]<='9'; g++ )
	{
		num *= 10;
		num += str[ g ] - '0';
	}

	if( neg )
	{
		num = -num;
	}

	return num;
}

void String::toLower()
{
	int g;

	for( g=0; g<len; g++ )
	{
		str[ g ] = tolower( str[ g ] );
	}
}

void String::toUpper()
{
	int g;

	for( g=0; g<len; g++ )
	{
		str[ g ] = toupper( str[ g ] );
	}
}

void String::write( FILE *out )
{
	fwrite( str, 1, len, out );
}

String::String()
{
	len = 0;
	size = 0;
	str = NULL;
}

String::~String()
{
	delete str;
}

struct HashNode
{
	int type;
	String first, second;
	HashNode *next;

	HashNode( int t, String& f, String& s, HashNode *n ) { type = t; first = f; second = s; next = n; }
	~HashNode() { delete next; }
};

class Hash
{
	HashNode *hash[ 1009 ];

public:
	int add( int t, String& f, String& s );
	HashNode *find( String& f );

	Hash();
	~Hash();
};

int Hash::add( int t, String& f, String& s )
{
	int num;
	HashNode *pos, *n;

	num = f.hash();
	pos = hash[ num ];
	while( pos!=NULL && pos->first!=f )
	{
		pos = pos->next;
	}

	if( pos!=NULL )
	{
		if( pos->type==NOD_FUNC || pos->type!=t )
		{
			fprintf( stderr, "You tried an invalid redefinition of %s!\n", f );
			exit( 69 );
		}
		pos->second = s;
		return 1;
	}

	n = new HashNode( t, f, s, hash[ num ] );
	if( n==NULL )
	{
		fprintf( stderr, "Out of memory!\n" );
		exit( 69 );
		return 1;
	}

	hash[ num ] = n;

	return 0;
}

HashNode *Hash::find( String& f )
{
	int num;
	HashNode *pos;

	num = f.hash();
	pos = hash[ num ];
	while( pos!=NULL && pos->first!=f )
	{
		pos = pos->next;
	}

	return pos;
}

Hash::Hash()
{
	for( int g=0; g<1009; g++ )
	{
		hash[ g ] = NULL;
	}
}

Hash::~Hash()
{
	for( int g=0; g<1009; g++ )
	{
		delete hash[ g ];
	}
}

struct Params
{
	String str[ 9 ];
	int num;

	Params() { num = 0; }
};

struct IO
{
	int type;

	int pos;
	FILE *file;
	String str;
	int line;

	int get();
	void operator+=( char );
	void operator+=( String& );

	IO() { type = TYP_NULL; }
	IO( FILE *f ) { type = TYP_FILE; line = 1; file = f; }
	IO( String& s ) { type = TYP_STR; pos = 0; str = s; }
	IO( char *s ) { type = TYP_STR; pos = 0; str = ""; }
};

int IO::get()
{
	int ch;

	switch( type )
	{
		case TYP_FILE:
			ch = fgetc( file );
			if( ch=='\n' ) line++;
			return ch;

		case TYP_STR:
			return str[ pos++ ];

		default:
			return 0;
	}
}

void IO::operator+=( char out )
{
	switch( type )
	{
		case TYP_FILE:
			fputc( out, file );
			break;

		case TYP_STR:
			str += char(out);
	}
}

void IO::operator+=( String& s )
{
	switch( type )
	{
		case TYP_FILE:
			s.write( file );
			break;

		case TYP_STR:
			str += s;
	}
}

class Interpret
{
	IO IONULL;
	Hash hash;
	IO g_in, g_out;
	int startLine;
	int debug;

	void error( char *s );
	void expect( char ch, char *s );
	int getToken( IO& in, String& str, char *s );
	void setDebug( int d ) { debug = d; }
	int null( IO& in, Params& p );
	int readExec( IO& in, IO& out );
	int literal( IO& in, IO& out );
	int doMacro( IO& in, IO& out, Params& p );
	int interpret( IO& in, IO& out, Params& p, int stopat );
	Macro *findMacro( String& tok );
	int getParams( IO& in, int ch, char *formal, Params& p, Params& pold );
	String& getVar( String& name );
	void setVar( String& name, String& val );
	void showDebug( char *s, Params& p );

public:
	Interpret( FILE *i, FILE *o, Params& p, int d );
};

void Interpret::error( char *s )
{
	fprintf( stderr, "Parse error: %s.\n", s );
	if( startLine!=0 )
	{
		fprintf( stderr, "Macro beginning at line: %i.\n", startLine );
		fprintf( stderr, "Current line: %i.\n", startLine, g_in.line );
	}
	exit( 69 );
}

void Interpret::expect( char ch, char *s )
{
	if( strchr( s, ch )==NULL )
	{
		fprintf( stderr, "Expected, one of: '%s'.\n", s );
		if( startLine!=0 )
		{
			fprintf( stderr, "Macro beginning at line: %i.\n", startLine );
			fprintf( stderr, "Current line: %i.\n", startLine, g_in.line );
			exit( 69 );
		}
	}
}

int Interpret::getToken( IO& in, String& str, char *s )
{
	int ch;
	int lit;

	lit = 0;
	for( ;; )
	{
		ch = in.get();
		if( lit==1 )
		{
			str += char(ch);
			lit = 0;
		}
		else
		{
			if( ch<=0 || strchr( s, ch )!=NULL )
			{
				break;
			}
			if( ch=='\\' )
			{
				lit = 1;
			}

			str += char(ch);
		}
	}

	return ch;
}

int Interpret::null( IO& in, Params& p )
{
	int ch;

	do
	{
		ch = interpret( in, IONULL, p, STO_ENDPARAM );
	}
	while( ch>0 && ch!='}' );

	return ch;
}

int Interpret::readExec( IO& in, IO& out )
{
	int open, lit;
	int ch;

	open = 0; lit = 0;

	for( ;; )
	{
		ch = in.get();
		if( ch<=0 ) break;
		if( lit==1 )
		{
			out += '\\';
			out += char( ch );
			lit = 0;
		}
		else
		{
			switch( ch )
			{
				case '\\':
					lit = 1;
					break;

				case '{':
					open++;
					out += char( ch );
					break;

				case '}':
					open--;
					if( open>=0 )
					{
						out += char( ch );
					}
					break;

				case ',':
					if( open==0 )
					{
						return ch;
					}
					out += char( ch );
					break;

				default:
					out += char( ch );
			}
			if( open==-1 )
			{
				break;
			}
		}
	}

	return ch;
}

int Interpret::literal( IO& in, IO& out )
{
	int open, lit;
	int ch;

	open = 0; lit = 0;

	for( ;; )
	{
		ch = in.get();
		if( ch<=0 ) break;
		if( lit==1 )
		{
			out += char( ch );
			lit = 0;
		}
		else
		{
			switch( ch )
			{
				case '\\':
					lit = 1;
					break;

				case '{':
					open++;
					out += char( ch );
					break;

				case '}':
					open--;
					if( open>=0 )
					{
						out += char( ch );
					}
					break;

				default:
					out += char( ch );
			}
			if( open==-1 )
			{
				break;
			}
		}
	}

	return ch;
}

int Interpret::doMacro( IO& in, IO& out, Params& p )
{
	String token;
	int ch;

	startLine = g_in.line;
	ch = in.get();
	if( ch<=0 )
	{
		error( "EOF in macro" );
	}
	if( ch=='?' )
	{
		HashNode *node;
		ch = getToken( in, token, ":}" );
		if( ch==':' )
		{
			ch = null( in, p );
		}
		out += getVar( token );
	}
	else if( ch=='!' )
	{
		IO temp( "" );
		ch = getToken( in, token, ":}" );

		while( ch==':' || ch==',' )
		{
			temp.str = ""; temp.pos = 0;
			ch = interpret( in, temp, p, STO_ENDPARAM );
		}

		setVar( token, temp.str );
	}
	else
	{
		token += ch;
		ch = getToken( in, token, ":}" );
		if( ch<=0 )
		{
			error( "End of macro name not found" );
		}

		token.toLower();

		if( token=="null" )
		{
			if( ch==':' )
			{
				ch = null( in, p );
			}
		}
		else if( token=="func" )
		{
			String token2;
			IO token3( "" );

			expect( ch, ":" );

			ch = getToken( in, token2, ",}" );
			expect( ch, "," );

			token2.toLower();

			ch = readExec( in, token3 );
			expect( ch, "}" );

			hash.add( NOD_FUNC, token2, token3.str );
		}
		else if( token=="eval" )
		{
			if( ch==':' )
			{
				IO temp( "" );
				IO temp2( "" );
				int c;

				do
				{
					temp.str = ""; temp.pos = 0;
					temp2.str = ""; temp2.pos = 0;
					ch = interpret( in, temp, p, STO_ENDPARAM );
					c = interpret( temp, temp2, p, STO_NULL );
				}
				while( ch>0 && ch!='}' );

				expect( ch, "}" );

				out += temp2.str;
			}
		}
		else if( token=="lit" )
		{
			ch = literal( in, out );
			expect( ch, "}" );
		}
		else
		{
			String str;
			Macro *macro;
			int num;
			Params params;
			int g, tempnum, tempnum2;
			IO temp( "" ), temp2( "" );

			macro = findMacro( token );
			if( macro==NULL )
			{
				HashNode *node;

				num = getParams( in, ch, "i*", params, p );
				node = hash.find( token );
				if( node==NULL || node!=NULL && node->type!=NOD_FUNC )
				{
					error( "Could not find function" );
				}

				temp.str << node->second;
				interpret( temp, out, params, STO_NULL );
				node->second << temp.str;
			}
			else
			{
				num = getParams( in, ch, macro->formal, params, p );
				if( debug )
				{
					showDebug( macro->name, params );
				}

				switch( macro->num )
				{
					case M_FOR:
						int count, first, last, inc;

						first = params.str[ 1 ].toInt();
						last = params.str[ 2 ].toInt();
						inc = params.str[ 3 ].toInt();

						count = first;
						temp.str << params.str[ 4 ];
						if( inc>0 && first<=last )
						{
							while( count<=last )
							{
								str = count;
								setVar( params.str[ 0 ], str );
								temp.pos = 0;
								interpret( temp, out, p, STO_NULL );
								count += inc;
							}
						}
						else if( inc<0 && first>=last )
						{
							while( count>=last )
							{
								str = count;
								setVar( params.str[ 0 ], str );
								temp.pos = 0;
								interpret( temp, out, p, STO_NULL );
								count += inc;
							}
						}
						params.str[ 4 ] << temp.str;
						break;

					case M_ADD:
						tempnum = 0;
						for( g=0; g<num; g++ )
						{
							tempnum += params.str[ g ].toInt();
						}
						str = tempnum;
						out += str;
						break;

					case M_SUB:
						tempnum = params.str[ 0 ].toInt();
						for( g=1; g<num; g++ )
						{
							tempnum -= params.str[ g ].toInt();
						}
						str = tempnum;
						out += str;
						break;

					case M_MUL:
						tempnum = params.str[ 0 ].toInt();
						for( g=1; g<num; g++ )
						{
							tempnum *= params.str[ g ].toInt();
						}
						str = tempnum;
						out += str;
						break;

					case M_DIV:
						tempnum = params.str[ 0 ].toInt();
						for( g=1; g<num; g++ )
						{
							tempnum /= params.str[ g ].toInt();
						}
						str = tempnum;
						out += str;
						break;

					case M_MOD:
						tempnum = params.str[ 0 ].toInt();
						for( g=1; g<num; g++ )
						{
							tempnum %= params.str[ g ].toInt();
						}
						str = tempnum;
						out += str;
						break;

					case M_UP:
						params.str[ 0 ].toUpper();
						out += params.str[ 0 ];
						break;

					case M_LOW:
						params.str[ 0 ].toLower();
						out += params.str[ 0 ];
						break;

					case M_LEN:
						str = params.str[ 0 ].getLength();
						out += str;
						break;

					case M_DNL:
						do
						{
							ch = in.get();
						}
						while( ch>0 && ch!='\n' );
						ch = '}';
						break;

					case M_HEX:
						str.setHex( params.str[ 0 ].toInt() );
						out += str;
						break;

					case M_HEXBYTE:
						str.setHexByte( params.str[ 0 ].toInt() );
						out += str;
						break;

					case M_IF:
						if( params.str[ 0 ].toInt() )
						{
							temp.str << params.str[ 1 ];
							interpret( temp, out, p, STO_NULL );
							params.str[ 1 ] << temp.str;
						}
						else
						{
							temp.str << params.str[ 2 ];
							interpret( temp, out, p, STO_NULL );
							params.str[ 2 ] << temp.str;
						}
						break;

					case M_OR:
						tempnum = 0;
						for( g=0; g<num && tempnum==0; g++ )
						{
							temp.pos = 0;
							temp.str << params.str[ g ];
							interpret( temp, temp2, p, STO_NULL );
							params.str[ g ] << temp.str;
							str << temp2.str;
							tempnum = str.toInt();
						}
						str = tempnum ? 1 : 0;
						out += str;
						break;

					case M_AND:
						tempnum = 1;
						for( g=0; g<num && tempnum; g++ )
						{
							temp.pos = 0;
							temp.str << params.str[ g ];
							interpret( temp, temp2, p, STO_NULL );
							params.str[ g ] << temp.str;
							str << temp2.str;
							tempnum = str.toInt();
						}
						str = tempnum ? 1 : 0;
						out += str;
						break;

					case M_EQ:
						str = params.str[ 0 ].smartCompare( params.str[ 1 ] )==0;
						out += str;
						break;

					case M_NE:
						str = params.str[ 0 ].smartCompare( params.str[ 1 ] )!=0;
						out += str;
						break;

					case M_GT:
						str = params.str[ 0 ].smartCompare( params.str[ 1 ] )>0;
						out += str;
						break;

					case M_GE:
						str = params.str[ 0 ].smartCompare( params.str[ 1 ] )>=0;
						out += str;
						break;

					case M_LT:
						str = params.str[ 0 ].smartCompare( params.str[ 1 ] )<0;
						out += str;
						break;

					case M_LE:
						str = params.str[ 0 ].smartCompare( params.str[ 1 ] )<=0;
						out += str;
						break;

					case M_NOT:
						str = !params.str[ 0 ].toInt();
						out += str;
						break;

					case M_XOR:
						tempnum = params.str[ 0 ].toInt();
						tempnum2 = params.str[ 1 ].toInt();
						str = tempnum && !tempnum2 || !tempnum && tempnum2;
						out += str;
						break;

					case M_NL:
						out += '\n';
						break;

					case M_RAND:
						str = rand();
						out += str;
						break;

					case M_MKLIST:
						str = params.str[ 0 ];
						for( g=1; g<num; g++ )
						{
							str += '\n';
							str += params.str[ g ];
						}
						out += str;
						break;

					case M_SUBSTR:
						str.substr( params.str[ 0 ], params.str[ 1 ].toInt(), params.str[ 2 ].toInt() );
						out += str;
						break;
				}
			}
		}
	}

	if( ch<=0 )
	{
		error( "End of file" );
	}

	startLine = 0;

	return ch;
}

int Interpret::interpret( IO& in, IO& out, Params& p, int stopat )
{
	int pos;
	int ch, lit;

	lit = 0;
	for( ;; )
	{
		ch = in.get();
		if( ch<=0 ) break;
		if( lit==1 )
		{
			out += char( ch );
			lit = 0;
		}
		else
		{
			switch( ch )
			{
				case ',':
				case '}':
					if( stopat==STO_ENDPARAM )
					{
						return ch;
					}
					else
					{
						out += char( ch );
					}
					break;

				case '\\':
					lit = 1;
					break;

				case '{':
					ch = doMacro( in, out, p );
					break;

				case '$':
					ch = in.get();
					if( ch>='1' && ch<='9' )
					{
						out += p.str[ ch-'1' ];
					}
					else
					{
						out += '$';
						out += char( ch );
					}
					break;

				default:
					out += char( ch );
			}
		}
	}

	return ch;
}

Macro *Interpret::findMacro( String& tok )
{
	int g;
	for( g=0; macros[ g ].num!=M_NULL; g++ )
	{
		if( tok==macros[ g ].name )
		{
			return &macros[ g ];
		}
	}

	return NULL;
}

int Interpret::getParams( IO& in, int ch, char *formal, Params& p, Params& pold )
{
	IO tempout( "" );
	int num;

	num = 0;
	if( ch==':' )
	{
		while( *formal!='\0' && num<9 && ch>0 && ch!='}' && ch>0 )
		{
			if( *formal=='+' ) formal--;
			switch( *formal )
			{
				case 'n':
					ch = null( in, p );
					break;

				case 'i':
					ch = interpret( in, tempout, pold, STO_ENDPARAM );
					p.str[ num++ ] << tempout.str;
					break;

				case 'e':
				case '?':
					ch = readExec( in, tempout );
					p.str[ num++ ] << tempout.str;
					break;

				case 't':
					ch = getToken( in, p.str[ num++ ], ",}" );
					break;
			}
			formal++;
			if( *formal=='*' ) formal--;
		}
	}

	if( *formal=='\0' || formal[ 1 ]!='*' )
	{
		if( ch!='}' || *formal!='?' && *formal!='\0' && *formal!='+' )
		{
			error( "Cannot find end of formal parameter list" );
		}
	}

	p.num = num;

	return num;
}

String& Interpret::getVar( String& name )
{
	HashNode *node;
	static String NULLSTR;

	node = hash.find( name );
	if( node!=NULL && node->type==NOD_VAR )
	{
		return node->second;
	}

	return NULLSTR;
}

void Interpret::setVar( String& name, String& val )
{
	hash.add( NOD_VAR, name, val );
}

void Interpret::showDebug( char *s, Params& p )
{
	int g;

	fprintf( stderr, "Doing: '%s'.", s );
	if( p.num!=0 )
	{
		fprintf( stderr, "  Params: " );
		for( g=0; g<p.num; g++ )
		{
			fprintf( stderr, "\"" );
			p.str[ g ].write( stderr );
			fprintf( stderr, "\"" );
			if( g!=p.num-1 )
			{
				fprintf( stderr, ", " );
			}
		}
	}
	fprintf( stderr, "\n" );
}

Interpret::Interpret( FILE *i, FILE *o, Params& p, int d )
	: g_in( i ), g_out( o )
{
	startLine = 0;
	debug = d;
	interpret( g_in, g_out, p, STO_NULL );
}

int main( int argc, char *argv[] )
{
	int c;

	int debug;

	debug = 0;

	for( ;; )
	{
		c = getopt( argc, argv, "dh" );
		if( c==-1 )
		{
			break;
		}

		switch( c )
		{
			case 0:
				fprintf( stderr, "Bad Option!\n" );
				exit( 69 );
				break;

			case 'd':
				debug = 1;
				break;

			case 'h':
				fprintf( stderr, "Stuff...\n" );
				exit( 69 );
				break;

			case '?':
				fprintf( stderr, "Type 'mpl --help' to learn how to use.\n" );
				exit( 69 );
				break;
		}
	}

	Params params;

	Interpret interpret( stdin, stdout, params, debug );

	return 0;
}
